from zope.interface import implements

from Acquisition import aq_base
from Globals import DevelopmentMode
from Products.Five.browser import BrowserView

from plone.reload.code import reload_code
from plone.reload.interfaces import IReload
from plone.reload.zcml import reload_zcml


class Reload(BrowserView):
    """Reload view.
    """
    implements(IReload)

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)
        self.message = None

    def __call__(self):
        if self.available():
            action = self.request.form.get('action')
            if action is not None:
                if action == 'code':
                    self.message = self.code_reload()
                elif action == 'zcml':
                    self.message = self.zcml_reload()
        return self.index()

    def available(self):
        if DevelopmentMode:
            return True
        return False

    def status(self):
        return self.message

    def code_reload(self):
        if not self.available():
            return

        reloaded = reload_code()

        result = ''
        if reloaded:
            result += 'Code reloaded:\n\n'
            result += '\n'.join(reloaded)
        else:
            result = 'No code reloaded!'
        return result

    def zcml_reload(self):
        if not self.available():
            return

        # We always do an implicit code reload so we can register all newly
        # added classes.
        reloaded = reload_code()
        reload_zcml()

        # TODO Minimize all caches, we only really want to invalidate the
        # local site manager from all caches
        aq_base(self.context)._p_jar.db().cacheMinimize()
        result = ''
        if reloaded:
            result += 'Code reloaded:\n\n'
            result += '\n'.join(reloaded)
        else:
            result = 'No code reloaded!'
        result += '\n\nGlobal ZCML reloaded.'
        return result



# xreload.py.

"""Alternative to reload().

This works by executing the module in a scratch namespace, and then
patching classes, methods and functions. This avoids the need to
patch instances. New objects are copied into the target namespace.

Taken and extended from xreload as posted by Guido van Rossum:

    http://mail.python.org/pipermail/edu-sig/2007-February/007787.html

"""

import marshal
import imp
import sys
import types
import inspect

import ExtensionClass

import zope.component

class Reloader(object):
    """Reload a module in place, updating classes, methods and functions.

    Args:
      mod: a module object

    Returns:
      The (updated) input object itself.
    """

    def __init__(self, module):
        self.mod = module

    def reload(self):
        # Get the module name, e.g. 'foo.bar.whatever'
        modname = self.mod.__name__

        # Get the module namespace (dict) early; this is part of the type check
        modns = self.mod.__dict__
        # Parse it into package name and module name, e.g. 'foo.bar' and 
        # 'whatever'
        i = modname.rfind(".")
        if i >= 0:
            pkgname, modname = modname[:i], modname[i+1:]
        else:
            pkgname = None
        # Compute the search path
        if pkgname:
            # We're not reloading the package, only the module in it
            pkg = sys.modules[pkgname]
            path = pkg.__path__  # Search inside the package
        else:
            # Search the top-level module path
            pkg  = None
            path = None  # Make find_module() uses the default search path
        # Find the module; may raise ImportError
        (stream, filename, (suffix, mode, kind)) = imp.find_module(modname, path)
        # Turn it into a code object
        try:
            # Is it Python source code or byte code read from a file?
            # XXX Could handle frozen modules, zip-import modules
            if kind not in (imp.PY_COMPILED, imp.PY_SOURCE):
                # Fall back to built-in reload()
                return reload(self.mod)
            if kind == imp.PY_SOURCE:
                source = stream.read()
                # PeterB: if we don't strip the source code and add newline we get 
                # a SyntaxError even if `python $filename` is perfectly happy.
                source = source.strip()+'\n'
                code = compile(source, filename, "exec")
            else:
                code = marshal.load(stream)
        finally:
            if stream:
                stream.close()
        # Execute the code im a temporary namespace; if this fails, no changes
        tmpns = {'__name__': '%s.%s' % (pkgname, modname),
                 '__file__': filename,
                 '__doc__': modns['__doc__']}
        exec(code, tmpns)
        # Now we get to the hard part
        oldnames = set(modns)
        newnames = set(tmpns)
        # Add newly introduced names
        for name in newnames - oldnames:
            modns[name] = tmpns[name]
        # Delete names that are no longer current
        for name in oldnames - newnames - set(["__name__"]):
            del modns[name]
        # Now update the rest in place
        for name in oldnames & newnames:
            modns[name] = self._update(modns[name], tmpns[name])
        # Done!
        return self.mod

    def _update(self, oldobj, newobj):
        """Update oldobj, if possible in place, with newobj.

        If oldobj is immutable, this simply returns newobj.

        Args:
          oldobj: the object to be updated
          newobj: the object used as the source for the update

        Returns:
          either oldobj, updated in place, or newobj.
        """
        if type(oldobj) is not type(newobj):
            # Cop-out: if the type changed, give up
            return newobj
        if hasattr(newobj, "__reload_update__"):
            # Provide a hook for updating
            return newobj.__reload_update__(oldobj)

        new_module = getattr(newobj, '__module__', None)
        if new_module != self.mod.__name__:
            # Do not update objects in-place that have been imported.
            # Just update their references.
            return newobj
        
        if isinstance(newobj, zope.interface.interface.Specification):
            # XXX we can't update interfaces because their internal
            # data structures break. We'll have to implement the reload method
            # for those and patch it in.
           return oldobj
        if inspect.isclass(newobj):
            return _update_class(oldobj, newobj)
        elif inspect.isfunction(newobj):
            return _update_function(oldobj, newobj)

        # XXX Support class methods, static methods, other decorators
        # Not something we recognize, just give up
        return newobj


def _update_function(oldfunc, newfunc):
    """Update a function object."""
    oldfunc.func_code = newfunc.func_code
    oldfunc.func_defaults = newfunc.func_defaults
    # XXX What else?
    return oldfunc


def _update_method(oldmeth, newmeth):
    """Update a method object."""
    # XXX What if im_func is not a function?
    _update_function(oldmeth.im_func, newmeth.im_func)
    return oldmeth


def _update_class(oldclass, newclass):
    """Update a class object."""
    # XXX What about __slots__?
    olddict = oldclass.__dict__
    newdict = newclass.__dict__
    oldnames = set(olddict)
    newnames = set(newdict)
    for name in newnames - oldnames:
        setattr(oldclass, name, newdict[name])
    for name in oldnames - newnames:
        delattr(oldclass, name)
    for name in oldnames & newnames - set(["__dict__", "__doc__"]):
        new = getattr(newclass, name)
        old = getattr(oldclass, name, None)
        if new == old:
            continue
        if old is None:
            setattr(oldclass, name, new)
            continue
        if isinstance(new, types.MethodType):
            _update_method(old, new)
        elif isinstance(new, types.FunctionType):
            # __init__ is a function
            _update_function(old, new)
        else:
            # Fallback to just replace the item
            setattr(oldclass, name, new)
    return oldclass



import os

from setuptools import setup, find_packages

name = 'plone.reload'
version = '0.9'

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

long_description=(
        read('README.txt')
        + '\n' +
        read('CHANGES.txt')
    )

setup(name=name,
      version=version,
      description="Configuration and code reload without Zope server restarts.",
      long_description=long_description,
      classifiers=[
        "Framework :: Plone",
        "Framework :: Zope2",
        "Programming Language :: Python",
        ],
      keywords='Zope Plone reload',
      author='Plone Foundation',
      author_email='plone-developers@lists.sourceforge.net',
      url='http://pypi.python.org/pypi/plone.reload',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['plone'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
      ],
      )
